@@ -78,6 +78,7 @@ struct _GVfsBackendSmb
   
   GMountSource *mount_source; /* Only used/set during mount */
   int mount_try;
+  gboolean mount_try_again;
   gboolean mount_cancelled;
 	
   gboolean password_in_keyring;
@@ -201,12 +202,16 @@ auth_callback (SMBCCTX *context,
       backend->user == NULL &&
       backend->domain == NULL)
     {
+      /* Try again if kerberos login fails */
+      backend->mount_try_again = TRUE;
       g_debug ("auth_callback - kerberos pass\n");
     }
   else if (backend->mount_try == 1 &&
            backend->user == NULL &&
            backend->domain == NULL)
     {
+      /* Try again if ccache login fails */
+      backend->mount_try_again = TRUE;
       g_debug ("auth_callback - ccache pass\n");
     }
   else
@@ -288,6 +293,9 @@ auth_callback (SMBCCTX *context,
 	    }
 	}
 
+      /* Try again if this fails */
+      backend->mount_try_again = TRUE;
+
       smbc_setOptionNoAutoAnonymousLogin (backend->smb_context,
                                           !anonymous);
 
@@ -484,6 +492,11 @@ do_mount (GVfsBackend *backend,
    */
   do
     {
+      /* The mount_try_again variable is here to avoid livelock in cases when
+       * EPERM is returned immediately without calling the auth_callback
+       * function. See: https://gitlab.gnome.org/GNOME/gvfs/-/issues/703.
+       */
+      op_backend->mount_try_again = FALSE;
       op_backend->mount_cancelled = FALSE;
 
       g_debug ("do_mount - try #%d \n", op_backend->mount_try);
@@ -523,7 +536,7 @@ do_mount (GVfsBackend *backend,
 
       op_backend->mount_try ++;
     }
-  while (TRUE);
+  while (op_backend->mount_try_again);
   
   g_free (uri);
   
@@ -2164,9 +2177,22 @@ do_move (GVfsBackend *backend,
 	  /* Unfortunately libsmbclient doesn't correctly return EXDEV, but falls back
 	     to EINVAL, so we try to guess when this happens: */
 	  (errsv == EINVAL && source_is_dir))
-	g_vfs_job_failed (G_VFS_JOB (job), 
-			  G_IO_ERROR, G_IO_ERROR_WOULD_RECURSE,
-			  _("Can’t recursively move directory"));
+        {
+          if (source_is_dir)
+            {
+              g_vfs_job_failed (G_VFS_JOB (job),
+                                G_IO_ERROR,
+                                G_IO_ERROR_WOULD_RECURSE,
+                                _("Can’t recursively move directory"));
+            }
+          else
+            {
+              g_vfs_job_failed (G_VFS_JOB (job),
+                                G_IO_ERROR,
+                                G_IO_ERROR_NOT_SUPPORTED,
+                                _("Operation not supported"));
+            }
+        }
       else
 	g_vfs_job_failed_from_errno (G_VFS_JOB (job), errsv);
     }
